import argparse
import os

from . import projection, tag
from .exif import ExifReader, TagReader


def rectify(args):
    if args.dir and not os.path.exists(args.dir):
        os.makedirs(args.dir)

    for src in args.image:
        if args.dir:
            dst = os.path.join(args.dir, os.path.basename(src.name))
        else:
            base, ext = os.path.splitext(src.name)
            dst = base + '_rectified' + ext

        rectified = projection.rectify(src, args.compass)

        params = {}
        if args.exif:
            params['exif'] = rectified.info['exif']

        with open(dst, 'wb') as fp:
            fp = projection.NonJFIFHeaderFile(fp)
            rectified.save(fp, 'JPEG', **params)

        src.close()

    return 0


def info(args):
    def formatter(reader, tags):
        for k, v in reader.items():
            if isinstance(v, TagReader):
                continue
            line = '0x{:04x}'.format(k)
            if k in tags:
                line += ' [{}]'.format(tags[k])
            if isinstance(v, tuple):
                v = '(' + ', '.join(map(str, v)) + ')'
            elif isinstance(v, bytes):
                try:
                    v = v.decode('ascii')
                except UnicodeDecodeError:
                    pass
            line += ': {}'.format(v)
            print(line)

    try:
        reader = ExifReader(args.image)
        makernote = reader.makernote
        theta = reader.theta
    except ValueError as e:
        print('Error:', e)
        return 1

    print('RICOH Marker Note')
    formatter(makernote, tag.MARKERNOTE_TAGS)
    print()
    print('THETA Subdir')
    formatter(theta, tag.THETASUBDIR_TGAS)

    return 0


def parse(argv=None):
    parser = argparse.ArgumentParser(description='THETA Image Tool')
    subparsers = parser.add_subparsers()
    subparsers.required = True
    subparsers.dest = 'command'

    # Rectify
    parser_rectify = subparsers.add_parser('rectify', help='rectify image')
    parser_rectify.set_defaults(func=rectify)
    parser_rectify.add_argument('image',
                                nargs='+',
                                type=argparse.FileType('rb'),
                                help='path to image')
    parser_rectify.add_argument('-c',
                                '--compass',
                                action='store_true',
                                help='use compass')
    parser_rectify.add_argument(
        '-d', '--dir', help='output directory (default: source directory)')
    parser_rectify.add_argument('-e',
                                '--exif',
                                action='store_true',
                                help='write EXIF')

    # Info
    parser_info = subparsers.add_parser('info',
                                        description='display THETA EXIF tag')
    parser_info.set_defaults(func=info)
    parser_info.add_argument('image',
                             type=argparse.FileType('rb'),
                             help='path to image')

    args = parser.parse_args(argv)
    return args.func(args)



# References:
# - http://xanxys.hatenablog.jp/entry/20131110/1384094832
# - http://www.ozhiker.com/electronics/pjmt/jpeg_info/ricoh_mn.html
# - http://www.sno.phy.queensu.ca/~phil/exiftool/TagNames/Ricoh.html
# - https://github.com/atotto/ricoh-theta-tools

import collections
import fractions
import io
import struct

from PIL import Image

from . import tag


class Handler(object):
    TABLE = {
        1: 'B',
        2: 'c',
        3: 'H',
        4: 'I',
        5: 'II',
        6: 'b',
        7: 'c',
        8: 'h',
        9: 'i',
        10: 'ii',
        11: 'f',
        12: 'd',
    }

    def __init__(self, typeid, formatter):
        self.typeid = typeid
        self.formatter = formatter

    @property
    def size(self):
        return self.formatter.size

    def read(self, fp):
        unpacked = self.formatter.unpack(fp.read(self.size))
        if len(unpacked) == 1:
            return unpacked[0]
        else:
            return fractions.Fraction(unpacked[0], unpacked[1])

    def write(self, fp, value):
        if self.typeid in (5, 10):
            value = fractions.Fraction(value)
            packed = self.formatter.pack(value.numerator, value.denominator)
        else:
            packed = self.formatter.pack(value)
        fp.write(packed)

    @classmethod
    def build_handler(cls, endian=''):
        handlers = {}
        for typeid, fmt in cls.TABLE.items():
            formatter = struct.Struct(endian + fmt)
            handlers[typeid] = cls(typeid, formatter)
        return handlers


Handler.lehandlers = Handler.build_handler('<')
Handler.behandlers = Handler.build_handler('>')


class TIFFHeader(object):
    def __init__(self, fp):
        self.endian = fp.read(2)
        if self.endian not in (b'II', b'MM'):
            raise ValueError('endian must be II or MM: {}'.format(self.endian))
        tiff_code = self.u16(fp)
        if tiff_code != 0x002A:
            raise ValueError('Invalid TIFF header.')
        self.zeroth_ifd_offset = self.u32(fp)
        fp.seek(self.zeroth_ifd_offset)

    @property
    def handlers(self):
        if self.endian == b'II':
            return Handler.lehandlers
        elif self.endian == b'MM':
            return Handler.behandlers

    @property
    def u16(self):
        return self.handlers[3].read

    @property
    def u32(self):
        return self.handlers[4].read


class TagReader(collections.MutableMapping):
    '''
    IFD reader class.
    '''
    SUBDIR_HEADER = {
        tag.EXIF_IFD_POINTER: 0,
        tag.GPS_INFO_IFD_POINTER: 0,
        tag.RICOH_SUBDIR: 20,
        tag.THETA_SUBDIR: 0
    }

    def __init__(self, fp, header=None):
        self.fp = fp
        self.tags = {}
        self.data = {}
        if header is None:
            self.header = TIFFHeader(fp)
        else:
            self.header = header

        for i in range(header.u16(fp)):
            tagid = header.u16(fp)
            tagtype = header.u16(fp)
            if tagid == 0 and tagtype == 0:
                continue
            try:
                handler = header.handlers[tagtype]
            except KeyError:
                raise ValueError('Invalid data type.')
            num = header.u32(fp)
            if handler.size * num > 4:
                offset = header.u32(fp)
            else:
                offset = fp.tell()
                fp.seek(4, io.SEEK_CUR)

            self.tags[tagid] = (handler, num, offset)

        self.nextifd_offset = header.u32(fp)

    def __str__(self):
        return str(self.asdict())

    def __getitem__(self, key):
        try:
            return self.data[key]
        except KeyError:
            handler, num, offset = self.tags[key]
            self.fp.seek(offset)
            value = tuple(handler.read(self.fp) for _ in range(num))
            if handler.typeid == 2:
                value = b''.join(value[:-1])
            elif handler.typeid == 7:
                value = b''.join(value)
            elif len(value) == 1:
                value = value[0]

            if key in self.SUBDIR_HEADER:
                self.fp.seek(value + self.SUBDIR_HEADER[key])
                self.data[key] = TagReader(self.fp, self.header)
            else:
                self.data[key] = value
            return self.data[key]

    def __setitem__(self, key, values):
        handler, num, offset = self.tags[key]
        if not isinstance(values, collections.abc.Sized):
            values = (values, )
        if num != len(values):
            raise ValueError('Invalid length of values.')

        self.fp.seek(offset)
        for value in values:
            handler.write(self.fp, value)

        if key in self.data:
            del self.data[key]

    def __delitem__(self, key):
        raise NotImplementedError

    def __iter__(self):
        return iter(self.tags)

    def __len__(self):
        return len(self.tags)

    def getoffset(self, key):
        handler, num, offset = self.tags[key]
        return offset

    def asdict(self):
        '''
        Convert to dictionary recursively.
        '''
        return self._dict(self)

    def _dict(self, obj):
        d = {}
        for k, v in obj.items():
            if isinstance(v, self.__class__):
                d[k] = self._dict(v)
            else:
                d[k] = v
        return d


class ExifReader(object):
    """EXIF reader class for THETA image."""

    EXIF_ID_CODE = b'Exif\x00\x00'
    RICOH_MAKERNOTE_CODE = b'Ricoh\x00\x00\x00'

    def __init__(self, img):
        if not isinstance(img, Image.Image):
            img = Image.open(img)
        if 'exif' not in img.info:
            raise ValueError('No EXIF.')

        self.img = img
        body = img.info['exif'][len(ExifReader.EXIF_ID_CODE):]
        self.fp = io.BytesIO(body)
        header = TIFFHeader(self.fp)

        self.ifdlist = []
        while True:
            ifd = TagReader(self.fp, header)
            self.ifdlist.append(ifd)
            if ifd.nextifd_offset:
                self.fp.seek(ifd.nextifd_offset)
            else:
                break

        self._makernote = None

    @property
    def exif(self):
        return self.ifdlist[0][tag.EXIF_IFD_POINTER]

    @property
    def gps(self):
        return self.ifdlist[0][tag.GPS_INFO_IFD_POINTER]

    @property
    def makernote(self):
        if not self._makernote:
            if tag.MAKER_NOTE not in self.exif:
                raise ValueError('No MakerNote.')

            self.fp.seek(self.exif.getoffset(tag.MAKER_NOTE))
            code_len = len(ExifReader.RICOH_MAKERNOTE_CODE)
            if self.fp.read(code_len) != self.RICOH_MAKERNOTE_CODE:
                raise ValueError('No RICOH maker note.')

            self._makernote = TagReader(self.fp, self.exif.header)
        return self._makernote

    @property
    def theta(self):
        if tag.THETA_SUBDIR not in self.makernote:
            raise ValueError('No THETA subdir.')
        return self.makernote[tag.THETA_SUBDIR]

    @property
    def thumbnail(self):
        offset = self.ifdlist[1][tag.JPEG_INTERCHANGE_FORMAT]
        length = self.ifdlist[1][tag.JPEG_INTERCHANGE_FORMAT_LENGTH]
        self.fp.seek(offset)
        fp = io.BytesIO(self.fp.read(length))
        return Image.open(fp)

    @thumbnail.setter
    def thumbnail(self, img):
        offset = self.ifdlist[1][tag.JPEG_INTERCHANGE_FORMAT]
        self.fp.seek(offset)
        img.save(self.fp, 'JPEG')
        end = self.fp.tell()
        self.fp.truncate()
        self.ifdlist[1][tag.JPEG_INTERCHANGE_FORMAT_LENGTH] = end - offset

    def tobytes(self):
        self.fp.seek(0)
        return self.EXIF_ID_CODE + self.fp.read()



EXIF_TAGS = {
    0x0103: 'Compression',
    0x011a: 'XResolution',
    0x011b: 'YResolution',
    0x0128: 'ResolutionUnit',
    0x0201: 'JPEGInterchangeFormat',
    0x0202: 'JPEGInterchangeFormatLength',
    0x8769: 'ExifIFDPointer',
    0x8825: 'GPSInfoIFDPointer',
    0x927c: 'MakerNote',
}

GPS_TAGS = {
    0x0011: 'GPSImgDirection',
}

MARKERNOTE_TAGS = {
    0x0001: 'MakerNoteType',
    0x0002: 'FirmwareVersion',
    0x0005: 'SerialNumber',
    0x1000: 'RecordingFormat',
    0x2001: 'RicohSubdir',
    0x4001: 'ThetaSubdir',
}

THETASUBDIR_TGAS = {
    0x0001: 'RicohCameraType',
    0x0002: 'HDRType',
    0x0003: 'ZenithEs',
    0x0004: 'CompassEs',
    0x0005: 'AbnormalAcceleration',
    0x0006: 'Zenith',
    0x0007: 'Compress',
    0x0101: 'FilmISO',
    0x0102: 'Aperture',
    0x0103: 'ExposureTime',
    0x0104: 'SensorSerial1',
    0x0105: 'SensorSerial2',
}


def _register(obj):
    import re
    import sys
    for tag, name in obj.items():
        name = re.sub('((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))', r'_\1', name)
        setattr(sys.modules[__name__], name.upper(), tag)


_register(EXIF_TAGS)
_register(GPS_TAGS)
_register(MARKERNOTE_TAGS)
_register(THETASUBDIR_TGAS)



from setuptools import find_packages, setup

setup(
    name='thetaexif',
    version='0.2',
    author='Regen',
    author_email='git@exadge.com',
    description='THETA EXIF Library',
    long_description=(open('README.rst').read() + '\n\n' +
                      open('CHANGES.rst').read()),
    url='https://github.com/regen100/thetaexif',
    platforms='any',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Utilities',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    packages=find_packages(exclude=['*.tests']),
    test_suite='thetaexif.tests',
    install_requires=['numpy', 'scipy', 'pillow'],
    entry_points={
        'console_scripts': ['theta-tool = thetaexif.cli:parse'],
    },
)



import functools

import numpy as np
from PIL import Image

from . import tag
from .exif import ExifReader


def getcoordinates(w, h, r):
    uu, vv = np.meshgrid(np.arange(w), np.arange(h))
    t = (uu - w / 2) * 2 * np.pi / w
    p = (vv - h / 2) * np.pi / h
    st, ct = np.sin(t), np.cos(t)
    sp, cp = np.sin(p), np.cos(p)
    xyz = np.dstack((cp * st, sp, cp * ct))
    xyz = np.einsum('ij,...j', r, xyz)
    t = np.arctan2(xyz[:, :, 0], xyz[:, :, 2])
    p = np.arcsin(xyz[:, :, 1])
    uu = t * w / 2 / np.pi + w / 2
    vv = p * h / np.pi + h / 2
    return np.dstack((vv, uu)).transpose(2, 0, 1)


def remap(imgarray, coordinates):
    from scipy import ndimage
    remapped = np.empty_like(imgarray)
    for c in range(imgarray.shape[2]):
        ndimage.map_coordinates(imgarray[..., c],
                                coordinates,
                                order=1,
                                mode='wrap',
                                output=remapped[..., c])

    return remapped


def rotation(axis, angle):
    axis = np.array(axis) / np.sqrt(np.dot(axis, axis))
    a = np.cos(angle / 2)
    b, c, d = axis * np.sin(angle / 2)
    return np.array([[
        a * a + b * b - c * c - d * d, 2 * (b * c - a * d), 2 * (b * d + a * c)
    ], [
        2 * (b * c + a * d), a * a + c * c - b * b - d * d, 2 * (c * d - a * b)
    ], [
        2 * (b * d - a * c), 2 * (c * d + a * b), a * a + d * d - b * b - c * c
    ]])


rx = functools.partial(rotation, [1, 0, 0])
ry = functools.partial(rotation, [0, 1, 0])
rz = functools.partial(rotation, [0, 0, 1])


def getpose(reader, compass=False):
    zenith = tuple(map(float, reader.theta[tag.ZENITH_ES]))
    z, x = np.deg2rad(zenith)
    r = rx(x).dot(rz(z))
    if compass:
        y = np.deg2rad(float(reader.theta[tag.COMPASS_ES]))
        r = ry(y).dot(r)
    return r


def rectify(img, compass=False):
    if not isinstance(img, Image.Image):
        img = Image.open(img)

    reader = ExifReader(img)
    r = getpose(reader, compass).T

    imgarray = np.asarray(img)
    h, w = imgarray.shape[:2]

    coord = getcoordinates(w, h, r)
    rectified = remap(imgarray, coord)
    resultimg = Image.fromarray(rectified)

    # Rewrite gyroscope and compass data
    reader.theta[tag.ZENITH_ES] = (0, 0)
    if compass:
        reader.theta[tag.COMPASS_ES] = 0
        reader.gps[tag.GPS_IMG_DIRECTION] = 0

    # Rewrite thumbnail
    size = reader.thumbnail.size
    reader.thumbnail = resultimg.resize(size, Image.ANTIALIAS)

    resultimg.info['exif'] = reader.tobytes()

    return resultimg


class NonJFIFHeaderFile():
    def __init__(self, fp):
        self.fp = fp

    def write(self, b):
        if b[:4] == b'\xff\xd8\xff\xe0' and b[20:22] == b'\xff\xe1':
            b = b[:2] + b[20:]
        return self.fp.write(b)
